from typing import Iterable
import numpy as np
import math

from .utils._measure_input_validation import check_weights, check_p, check_sets_cardinality
from . import DISTANCE_HAMMING, DISTANCE_EUCLIDEAN, DISTANCE_NORMALIZED_HAMMING, DISTANCE_NORMALIZED_EUCLIDEAN
from . import WANGXIN_DISTANCE_1, WANGXIN_DISTANCE_2
from .sets import FuzzySet, IntuitionisticFuzzySet


def atanassov(A: FuzzySet, B: FuzzySet, distance_type: str = DISTANCE_HAMMING) -> np.float64:
    """ Distance proposed by K.T. Atanassov, from the related article: "Intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    distance_type: str, optional
        Type of computed distance: 

        >>> DISTANCE_HAMMING,
        >>> DISTANCE_EUCLIDEAN
        >>> DISTANCE_NORMALIZED_HAMMING
        >>> DISTANCE_NORMALIZED_EUCLIDEAN
        
    Returns
    -------
    numpy.float64
        The distance between the two sets provided.
    """
    check_sets_cardinality(A, B)
    if distance_type == DISTANCE_HAMMING:
        return (
            np.sum(
                np.absolute(A.membership_values - B.membership_values) + 
                np.absolute(A.non_membership_values - B.non_membership_values)
            )
            ) / 2.0

    elif distance_type == DISTANCE_EUCLIDEAN:
        return np.sqrt(
                np.sum(
                    np.power(A.membership_values - B.membership_values, 2.0) + 
                    np.power(A.non_membership_values - B.non_membership_values, 2.0)
                ) / 2.0
            )

    elif distance_type == DISTANCE_NORMALIZED_HAMMING:
        return np.sum(
                np.absolute(A.membership_values - B.membership_values) + 
                np.absolute(A.non_membership_values - B.non_membership_values)
            ) / (2.0 * len(A))

    elif distance_type == DISTANCE_NORMALIZED_EUCLIDEAN:
        return np.sqrt(
            np.sum(
                np.power(A.membership_values - B.membership_values, 2.0) + 
                np.power(A.non_membership_values - B.non_membership_values, 2.0)
            ) / (2.0 * len(A))
        )

    else:
        raise ValueError(
            "Invalid distance type provided. Please check the available flags.")


def szmidt_kacprzyk(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet, distance_type: str = DISTANCE_HAMMING) -> np.float64:
    """ Distances proposed by E. Szmidt and A. Kacprzyk, from the related article: "Distances between intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    distance_type: str, optional
        Type of computed distance: 
        
        >>> DISTANCE_HAMMING
        >>> DISTANCE_EUCLIDEAN
        >>> DISTANCE_NORMALIZED_HAMMING
        >>> DISTANCE_NORMALIZED_EUCLIDEAN
        
    Returns
    -------
    numpy.float64
        The distance between the two sets provided.
    """
    check_sets_cardinality(A, B)
    if distance_type == DISTANCE_HAMMING:
        return np.sum(
            np.absolute(A.membership_values - B.membership_values) +
            np.absolute(A.non_membership_values - B.non_membership_values) +
            np.absolute(A.hesitation_degrees - B.hesitation_degrees)
        ) / 2.0

    elif distance_type == DISTANCE_EUCLIDEAN:
        return np.sqrt(
            (
                np.sum(
                    np.power(A.membership_values - B.membership_values, 2.0) +
                    np.power(A.non_membership_values - B.non_membership_values, 2.0) +
                    np.power(A.hesitation_degrees - B.hesitation_degrees, 2.0))
            ) / 2.0
        )

    elif distance_type == DISTANCE_NORMALIZED_HAMMING:
        return np.sum(
            np.absolute(A.membership_values - B.membership_values) + 
            np.absolute(A.non_membership_values - B.non_membership_values) + 
            np.absolute(A.hesitation_degrees - B.hesitation_degrees)
        ) / (2.0 * len(A))

    elif distance_type == DISTANCE_NORMALIZED_EUCLIDEAN:
        return np.sqrt(
            np.sum(
                np.power(A.membership_values - B.membership_values, 2.0) + 
                np.power(A.non_membership_values - B.non_membership_values, 2.0) + 
                np.power(A.hesitation_degrees - B.hesitation_degrees, 2.0)
            ) / (2.0 * len(A))
        )

    else:
        raise ValueError(
            "Invalid distance type provided. Please check the available flags.")


def wang_xin(A: FuzzySet, B: FuzzySet, distance_type: int = WANGXIN_DISTANCE_1, weights: Iterable = None, p: int = 1) -> np.float64:
    """ Distances proposed by W. Wang and X. Xin, from the related article: 
        "Distance measure between intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        distance_type: int, optional
        Type of computed distance: 
    
        >>> WANGXIN_DISTANCE_1
        >>> WANGXIN_DISTANCE_2
    
    weights : list of floats
        List of weights for each membership/non-membership value.
    p : int
        Positive integer >= 1.
        
    Returns
    -------
    numpy.float64
        The distance between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p, distance_type, WANGXIN_DISTANCE_2)
    check_weights(weights, len(A), distance_type, WANGXIN_DISTANCE_1)
    n = len(A)

    m_diff = np.absolute(A.membership_values - B.membership_values)
    v_diff = np.absolute(A.non_membership_values - B.non_membership_values)

    if distance_type == WANGXIN_DISTANCE_1:
        if weights is None:
            return np.sum(
                (m_diff + v_diff) / 4.0 +
                np.maximum(m_diff, v_diff) / 2.0
            ) / n
        else:
            return np.sum(
                weights *
                ((m_diff + v_diff) / 4.0 +
                 np.maximum(m_diff, v_diff) / 2.0)
            ) / np.sum(weights)

    elif distance_type == WANGXIN_DISTANCE_2:
        multiplier = (1. / math.pow(n, 1. / p))
        summation = np.sum(
            np.power((m_diff + v_diff) / 2., p)
        )
        return multiplier * np.power(summation, (1.0 / p))


def yang_chiclana(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet,  distance_type: str = DISTANCE_HAMMING) -> np.float64:
    """ Distances proposed by Y. Yang and F. Chiclana, from the related article: "Consistency of 2D and 3D distances of intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    distance_type: str, optional
        Type of computed distance: 

        >>> DISTANCE_HAMMING
        >>> DISTANCE_EUCLIDEAN
        >>> DISTANCE_NORMALIZED_HAMMING or 
        >>> DISTANCE_NORMALIZED_EUCLIDEAN
        
    Returns
    -------
    numpy.float64
        The distance between the two sets provided.
    """
    check_sets_cardinality(A, B)
    n = len(A)
    delta_memberships = np.absolute(A.membership_values - B.membership_values)
    delta_non_memberships = np.absolute(A.non_membership_values - B.non_membership_values)
    delta_hesitations = np.absolute(A.hesitation_degrees - B.hesitation_degrees)

    if distance_type == DISTANCE_HAMMING:
        return np.sum(
            np.amax((delta_memberships, delta_non_memberships, delta_hesitations), axis=0)
        )
    elif distance_type == DISTANCE_EUCLIDEAN:
        return np.sqrt(
            np.sum(
                np.amax((delta_memberships ** 2, delta_non_memberships ** 2, delta_hesitations ** 2), axis=0)
            )
        )
    elif distance_type == DISTANCE_NORMALIZED_HAMMING:
        return np.sum(
            np.amax((delta_memberships, delta_non_memberships, delta_hesitations), axis=0)
        ) / n
    elif distance_type == DISTANCE_NORMALIZED_EUCLIDEAN:
        return np.sqrt(
            np.sum(
                np.amax((delta_memberships ** 2, delta_non_memberships ** 2, delta_hesitations ** 2), axis=0)
            ) / n
        )
    else:
        raise ValueError(
            "Invalid distance type provided. Please check the available flags.")


def grzegorzewski(A: FuzzySet, B: FuzzySet, distance_type: str = DISTANCE_HAMMING) -> np.float64:
    """ Distances proposed by P. Grzegorzewski from the related article: 
    "Distances between intuitionistic fuzzy sets and/or interval-valued fuzzy sets based on the Hausdorff metric"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    distance_type: str, optional
        Type of computed distance: 
        >>> DISTANCE_HAMMING
        >>> DISTANCE_EUCLIDEAN
        >>> DISTANCE_NORMALIZED_HAMMING** or 
        >>> DISTANCE_NORMALIZED_EUCLIDEAN**.
        
    Returns
    -------
    numpy.float64
        The distance between the two sets provided.
    """
    check_sets_cardinality(A, B)
    m_diff = A.membership_values - B.membership_values
    v_diff = A.non_membership_values - B.non_membership_values

    if distance_type == DISTANCE_HAMMING:
        return np.sum(np.maximum(np.absolute(m_diff), np.absolute(v_diff)))
    elif distance_type == DISTANCE_EUCLIDEAN:
        return np.sqrt(np.sum(np.maximum(m_diff ** 2, v_diff ** 2)))
    elif distance_type == DISTANCE_NORMALIZED_HAMMING:
        return np.sum(np.maximum(np.absolute(m_diff), np.absolute(v_diff))) / float(len(A))
    elif distance_type == DISTANCE_NORMALIZED_EUCLIDEAN:
        return np.sqrt(np.sum(np.maximum((m_diff) ** 2, v_diff ** 2)) / float(len(A)))
    else:
        raise ValueError(
            "Invalid distance type provided. Please check the available flags.")


def vlachos_sergiadis(A: FuzzySet, B: FuzzySet) -> np.float64:
    """ Distance proposed by I.K. Vlachos, G.D. Sergiadis from the related article: "Intuitionistic fuzzy information - Applications to pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        
    Returns
    -------
    numpy.float64
        The distance between the two sets provided.
    """
    check_sets_cardinality(A, B)
    def _log(arr):
        invalids = np.logical_or(
            arr <= 0, 
            np.isnan(arr), 
            np.isinf(arr)
        )
        invalid_indxs = np.where(invalids)
        valid_indxs = np.where(np.logical_not(invalids))
        arr[invalid_indxs] = 0
        arr[valid_indxs] = np.log(arr[valid_indxs])
        return arr

    def iifs(setA: FuzzySet, setB: FuzzySet):
        return np.sum(
            setA.membership_values *
            _log(
                setA.membership_values /
                (0.5 * (setA.membership_values + setB.membership_values))
            ) +
            setA.non_membership_values *
            _log(
                setA.non_membership_values /
                (0.5 * (setA.non_membership_values + setB.non_membership_values))
            )
        )
    with np.errstate(divide='ignore'):
        return iifs(A, B) + iifs(B, A)



import numpy as np
from typing import Union, Iterable

from ..sets import FuzzySet, IntuitionisticFuzzySet


def check_weights(weights: Iterable, set_cardinality: int, actual_measure_type: Union[Iterable, str, int]=None, *measure_types_required) -> np.ndarray:
    """ Input validation for measures that require weights.

    Checks if the weights have the same size as the sets' cardinality and if the values are [0, 1].
    Validation is performed only if there is no requirement for a specific measure type or if the measure type provided (actual_measure_type)
    requires the weights parameter.

        Args:
            weights: Input weights.
            set_cardinality: Cardinality (length) of the sets.
            actual_value: The type of measure that was provided.
            value_required: In case the weights are used for a specific measure type (and all).
        Returns:
            The converted weights.
        Raises:
            ValueError if weights size != set_cardinality or if weights values are not [0, 1].
    """
    if weights is None:
        return weights
    
    weights = np.array(weights)
    if len(measure_types_required) == 0 or actual_measure_type in measure_types_required:
        if weights.size != set_cardinality:
            raise ValueError(
                "Weight parameter must have the same size as sets A and B!({} vs {})".format(weights.size, n))

        outliers = np.where(np.logical_or(
            weights < 0, weights > 1))[0]
        outliers = weights[outliers]

        if len(outliers) > 0:
            raise ValueError(
                "Weight values must be [0, 1]. (found (some) {})".format(outliers[:5]))
    return weights


def check_p(p: Union[int, float], actual_measure_type: Union[Iterable, str, int] = None, *measure_types_required) -> None:
    """ Input validation for measures that require the parameter p.

    Checks if the p is an int and if it is >=1.

        Args:
            p: Input p.
            actual_value: The type of measure that was provided.
            value_required: In case the weights are used for a specific measure type (and all).
        Raises:
            ValueError if p is not an integer or if it is < 1.
    """
    if len(measure_types_required) == 0 or actual_measure_type in measure_types_required:
        if not np.issubdtype(type(p), int):
            raise ValueError(
                "p parameter must be an integer, not {}".format(type(p)))
        elif p < 1:
            raise ValueError(
                "p parameter must be >= 1, not {}".format(p))


def check_sets_cardinality(A: FuzzySet, B: FuzzySet) -> None:
    """ Checks if sets have the same cardinality

        Args:
            A: FuzzySet.
            B: FuzzySet.

        Raises:
            ValueError if the two sets have different cardinalities
    """
    validate_subset_sizes(A)
    validate_subset_sizes(B)

    if len(A) != len(B):
        raise ValueError("A and B sets must be have the same sizes.({} and {})".format(len(A), len(B)))


def validate_subset_sizes(set: FuzzySet) -> bool:
    """ Checks if set's values have the same sizes

        Args:
            A: FuzzySet.

        Raises:
            ValueError if the the set's values have different sizes.
    """
    sets_to_check = ["membership_values",
                     "non_membership_values", "hesitation_degrees"]

    error_msg = []
    for subset1_name in sets_to_check:
        for subset2_name in sets_to_check:
            if subset1_name == subset2_name:
                continue
            if not (hasattr(set, subset1_name) and hasattr(set, subset2_name)):
                continue

            values1 = getattr(set, subset1_name)
            values2 = getattr(set, subset2_name)
            assert len(values1) == len(values2), "{} and {} have different sizes! ({} and {})".format(
                subset1_name, subset2_name, len(values1), len(values2))
    validation_succeeded = len(error_msg) > 0
    return validation_succeeded, error_msg



from collections.abc import Callable
from typing import Iterable, List
import warnings
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.multiclass import unique_labels
import numpy as np

from ..sets import FuzzySet, IntuitionisticFuzzySet


def classify(class_patterns: Iterable[FuzzySet], sample_pattern: FuzzySet, 
            measure_caller: Callable, *, is_distance=True, return_confidence=False, **kwargs) -> np.intp:
    """ Simple classification method to classify a sample pattern given class patterns, using the measure provided.

    For each class pattern c, calculates the measure from measure_caller between c and sample_pattern.
    The class is chosen by the min/max measure between the sample and class patterns, depending on is_distance.
    If return_confidence is true, returns the degree of confidence for the chosen class.

    Parameters
    ----------
    class_patterns : list[FuzzySet]
        Class patterns to which the sample_pattern is classified.
    sample_pattern : FuzzySet
        The sample to be classified.
    measure_caller : Callable
        The measure function to use when measuring the sets of the ideally thresholded image and the current image set.
    is_distance : bool
        If the measure provided is a distance or a similarity. Used to pick the best measure calculated.
    return_confidence : bool
        Whether the confidence degree is returned. Can only be calculated if is_distance is True.
    **kwargs : additional arguments
        Passed to the measure_caller.
    
    Returns
    -------
    numpy.intp
        The class of sample_pattern.
    
    See Also
    --------
    fsmpy.utils.confidence_degree
    """
    if not is_distance and return_confidence:
        warnings.warn("Degree of Confidence can only be calculated for distances. Returning only prediction.")
        return_confidence = False
    measures = [
        measure_caller(sample_pattern, class_pattern, **kwargs) * (1 if is_distance else -1)
        for class_pattern in class_patterns
    ]
    prediction = np.argmin(measures)
    if return_confidence:
        from . import confidence_degree
        return prediction, confidence_degree(np.min(measures), measures[:prediction] + measures[prediction:])
    else:
        return prediction


class FuzzyTextClassifier(BaseEstimator, ClassifierMixin):
    """ Text document classification proposed by P. Intarapaiboon from the related article: 
    "Text classification using similarity measures on intuitionistic fuzzy sets".

    Train and classification object to classify text documents following
    the proposed method.

    Follows Estimator API from scikit-learn Estimator object.

    Attributes
    ----------
    measure_caller : Callable
        Measure to use during prediction process.
    measure_kwargs : dict
        Passed to the measure_caller.
    is_distance : bool
        If measure_caller is a distance or not.

    Methods
    -------
    fit : FuzzyTextClassifier
        "Trains" the classifier.
    predict : list[np.intp]
        List of predicted class for each sample
    predict_proba : list[list[np.float64]]
        List of the measure results between each class, for each sample.
    get_params : dict
        Parameters of the classifier.
    set_params : FuzzyTextClassifier
        Sets the parameters of the classifier.

    See also
    --------
    sklearn.base.BaseEstimator
    
    """
    def __init__(self, measure_caller: Callable, *, is_distance: bool = True, **measure_kwargs):
        """ Constructor method.
        
        Parameters
        ----------
        measure_caller : Callable
            The measure to be tested.
        is_distance : bool
            If the measure provided is a distance or a similarity. Used to pick the best measure calculated.
        **measure_kwargs : measure arguments
            Passed to the measure_caller.
        """
        self.measure_caller = measure_caller

        self._param_names = ["measure_caller"] + list(measure_kwargs.keys())
        self._measure_kwargs = measure_kwargs
        self.class_patterns = None
        self.is_distance = is_distance

    def fit(self, X: Iterable[FuzzySet], y: Iterable) -> object:
        """ "Trains" on the X data prpovided.
        
        Calculates the membership and non-membership values of each word for each unity class in y.

        Parameters
        ----------
        X : Iterable[FuzzySet]
            Data to train upon.
        y : Iterabale
            Target vector relative to X.
            
        Returns
        -------
        self: Object
            An instance of the estimator.

        See also
        --------
        fsmpy.utils.calculate_documents_membership
        """
        if not all(isinstance(x, FuzzySet) for x in X):
            raise TypeError(
                "Expected X to be an Iterable of types FuzzySet or IntuitionisticFuzzySet, got {}".format(type(X)))

        y = np.array(y)
        X = np.array(X)
        
        self.classes_ = unique_labels(y)

        self.class_patterns = {_class: None for _class in self.classes_}
        for _class in self.classes_:
            same_class_docs = y == _class
            
            mean_m = np.average(
                [x.membership_values for x in X[same_class_docs]], axis=0)
            mean_v = np.average(
                [x.non_membership_values for x in X[same_class_docs]], axis=0)
            if hasattr(X[0], "hesitation_degrees"):
                mean_h = np.average(
                    [x.hesitation_degrees for x in X[same_class_docs]], axis=0)
                self.class_patterns[_class] = IntuitionisticFuzzySet(
                    mean_m, mean_v, mean_h)
            else:
                self.class_patterns[_class] = FuzzySet(mean_m, mean_v)
        return self

    def predict(self, X: Iterable[FuzzySet]) -> np.ndarray:
        """ Predict class labels for samples X.
        
        Calculates the membership and non-membership values of each word for each unity class in y.

        Parameters
        ----------
        X : Iterable[FuzzySet]
            Samples.
            
        Returns
        -------
        list[np.intp]
            Predicted class label per sample.
        """
        if not all(isinstance(x, FuzzySet) for x in X):
            raise TypeError(
                "Expected X to be an Iterable of types FuzzySet or IntuitionisticFuzzySet, got {}".format(type(X)))

        predictions = []
        for x in X:
            predictions.append(
                np.argmin(
                    [
                        self.measure_caller(x, self.class_patterns[_class], **self._measure_kwargs) * (1 if self.is_distance else -1)
                        for _class in self.classes_
                    ]
                )
            )
        return np.array([self.classes_[pred] for pred in predictions])

    def predict_proba(self, X: Iterable[FuzzySet]) -> np.ndarray:
        """ Measures of each sample.
        
        The returned values are the results returned when the measure_caller is called
        for all classes for each label.

        Parameters
        ----------
        X : Iterable[FuzzySet]
            Samples.
            
        Returns
        -------
        list[list[float]]
            Returns the measures of the sample for each class, where classes are in self.classes_
        """
        if not all(isinstance(x, FuzzySet) for x in X):
            raise TypeError(
                "Expected X to be an Iterable of types FuzzySet or IntuitionisticFuzzySet, got {}".format(type(X)))

        probas = []
        for x in X:
            probas.append(
                [self.measure_caller(x, self.class_patterns[_class], **self._measure_kwargs)
                 for _class in self.classes_]
            )
        return np.vstack(probas)

    def get_params(self, deep=True) -> dict:
        params = {k: v for k, v in self._measure_kwargs.items()}
        params["measure_caller"] = self.measure_caller
        params["is_distance"] = self.is_distance
        return params

    def set_params(self, **parameters) -> object:
        for parameter, value in parameters.items():
            if parameter in ["is_distance", "measure_caller"]:
                setattr(self, parameter, value)
            else:
                self._measure_kwargs[parameter] = value
        return self



from typing import Callable, Tuple
import numpy as np

from ..sets import IntuitionisticFuzzySet


def threshold(image: np.ndarray, measure_caller: Callable, *, is_distance: bool = True, l: float = 0.2, **kwargs) -> Tuple[np.ndarray, bool]:
    """ Thresholds the input image following the proposed method by T. Chaira and A.K. Ray, from the related article:
    "Threshold selection using fuzzy set theory".

    For each thresold value t [0, 255], the set S1 of the ideally thresholded image is calculated and 
    the set S2 of thresholded image with value t.
    The selected threshold value is the one with the min/max (Depending on is_distance) measure
    between the sets for each threshold value t.

    Parameters
    ----------
    image : np.ndarray
        Single channel input image.
    measure_caller : Callable
        The measure function to use when measuring the sets of the ideally thresholded image and the current image set.
    is_distance : bool
        If the measure provided is a distance or a similarity. Used to pick the best measure calculated.
    l : float
        Used to calculate the membership values of the thresholded image with value t.
    **kwargs : additional arguments
        Passed to the measure_caller.
        
    Returns
    -------
    thresholded_image : np.ndarray
        Thresholded image.
    threshold_value : np.float64
        Threshold value.
    """
    x, y = image.shape
    img = image.copy().flatten()
    mx = np.zeros_like(img, dtype=np.float64)
    fmax = np.max(img)
    fmin = np.min(img)
    c = 1 / (fmax - fmin)

    measures = []
    for t in range(256):
        low_count = np.sum(img <= t)
        m0 = (low_count * t) / low_count
        if np.isnan(m0) or not np.isfinite(m0):
            m0 = 0
        high_count = np.sum(img > t)
        m1 = (high_count * t) / high_count
        if np.isnan(m1) or not np.isfinite(m1):
            m1 = 0

        np.putmask(mx, img >= t, np.exp(-c * np.abs(img - m0)))
        np.putmask(mx, img < t, np.exp(-c * np.abs(img - m1)))

        A = IntuitionisticFuzzySet(
            mx, 
            (1.0 - mx) - ((1.0 - mx) / (1.0 + l * mx))
        )
        B = IntuitionisticFuzzySet([1.0] * len(A))
        measures.append(
            measure_caller(A, B, **kwargs)
        )
    
    t = np.argmin(measures) if is_distance else np.argmax(measures)
    mask = np.zeros_like(img)
    mask[img >= t] = 255
    return mask.reshape((x, y)), t




from __future__ import annotations
from collections.abc import Iterable
import numpy as np


class FuzzySet:
    """ Object to represent a Fuzzy Sets.

    Attributes
    ----------
    membership_values : 1-d np.ndarray
        Membership values of the set.
    non_membership_values : 1-d np.ndarray
        Non-membership values of the set.

    Notes
    --------
    An important note as to how the Fuzzy Sets are actually represented.
    For three patterns denoted in Fuzzy Sets in a universe 
    :math:`U=\{u_1, u_2, u_3\}`
    
    mathematically, they are represented as follows :math:`(μ, ν)`
    """
    def __init__(self, membership_values : Iterable, non_membership_values : Iterable = None):
        """ Constructor method.

        For single valued sets, use an Iterable with a single value.

        Parameters
        ----------
        membership_values : 1-d Iterable
            Membership values of the set.
        non_membership_values : 1-d Iterable, optional
            Non-membership values of the set. If not provided, an np.array is 
            initialized with zeros with the same size as membership_values.
        
        """
        self.membership_values = np.array(membership_values)
        if non_membership_values is None:
            self.non_membership_values = np.zeros_like(self.membership_values)
        else:
            self.non_membership_values = np.array(non_membership_values)

    def __setattr__(self, name : str, value : Iterable):
        """ Sets the values of the specified attribute.
        
        Parameters
        ----------
        name : str
            Name of the attribute to set.
        value : Iterable
            The values to set to the attribute.
        
        Raises
        ------
        TypeError if value is not an Iterable or a numpy array
        """
        if not isinstance(value, Iterable):
            raise TypeError(
                "{} must be an Iterable or a numpy array, not {}!".format(name, type(value)))

        self.__dict__[name] = np.array(value)

    def __len__(self):
        """ Returns cardinality of the set.
        
        Returns the size of the membership_values.

        Returns
        -------
        int
        """
        return len(self.membership_values)

    def __str__(self):
        """ String representation of the set.
        
        Returns
        -------
        str
        """
        return "μ: {},\nv:{}".format(self.membership_values, self.non_membership_values)


class IntuitionisticFuzzySet(FuzzySet):
    """ Object to represent a Intuitionistic Fuzzy Sets. 

    The important difference from FuzzySet is that it incorporates the
    hesitation degree values.

    Attributes
    ----------
    membership_values : 1-d np.ndarray
        Membership values of the set.
    non_membership_values : 1-d np.ndarray
        Non-membership values of the set.
    hesitation_degrees : 1-d np.ndarray
        Hesitation degree values of the set.

    Notes
    --------
    An important note as to how the Intuitionistic Fuzzy Sets are actually represented.
    For three patterns denoted in Intuitionistic Fuzzy Sets in a universe 
    :math:`U=\{u_1, u_2, u_3\}`
    
    mathematically, they are represented as follows :math:`(μ, ν)`
    """
    def __init__(self, membership_values: Iterable, non_membership_values: Iterable = None, hesitation_degrees: Iterable = None):
        """ Constructor method.

        For single valued sets, use an Iterable with a single value.

        Parameters
        ----------
        membership_values : 1-d Iterable
            Membership values of the set.
        non_membership_values : 1-d Iterable, optional
            Non-membership values of the set. If not provided, an np.array is 
            initialized with zeros with the same size as membership_values.
        hesitation_degrees : 1-d Iterable, optional
            Hesitation degrees values of the set. If not provided, it is set to

            >>> hesitation_degrees = 1.0 - membership_values - self.non_membership_values
        
        """
        super().__init__(membership_values, non_membership_values)
        if hesitation_degrees is None:
            self.hesitation_degrees = 1.0 - np.array(membership_values) - np.array(self.non_membership_values)
        else:
            self.hesitation_degrees = hesitation_degrees

    def __len__(self):
        """ Returns cardinality of the set.
        
        Returns the size of the membership_values.

        Returns
        -------
        int
        """
        return len(self.membership_values)

    def __str__(self):
        """ String representation of the set.
        
        Returns
        -------
        str
        """
        return "μ: {}\nv:{}\nπ: {}".format(self.membership_values, self.non_membership_values, self.hesitation_degrees)



__version__ = "0.1.1"

""" Includes measure types variables, specifying the type of measure to use in specific measures.
The types are used to specify a specific measure function proposed in a study.
For example, Atanassov proposed 4 types of distances in a single study, the following
constants are used to specify each distance type to calculate using the `atanassov` measure:

>>> DISTANCE_HAMMING
>>> DISTANCE_EUCLIDEAN
>>> DISTANCE_NORMALIZED_HAMMING
>>> DISTANCE_NORMALIZED_EUCLIDEAN

Includes the following types:

>>> DISTANCE_HAMMING = 'Hamming'
>>> DISTANCE_EUCLIDEAN = 'Euclidean'
>>> DISTANCE_NORMALIZED_HAMMING = 'normalized Hamming'
>>> DISTANCE_NORMALIZED_EUCLIDEAN = 'normalized Euclidean'

>>> WANGXIN_DISTANCE_1 = 1
>>> WANGXIN_DISTANCE_2 = 2

>>> LIANG_SHI_SIMILARITY_1 = 'e'
>>> LIANG_SHI_SIMILARITY_2 = 's'
>>> LIANG_SHI_SIMILARITY_3 = 'h'

>>> HUNG_YANG_1_SIMILARITY_1 = 'l'
>>> HUNG_YANG_1_SIMILARITY_2 = 'e'
>>> HUNG_YANG_1_SIMILARITY_3 = 'c'

>>> HUNG_YANG_2_SIMILARITY_1 = 'l'
>>> HUNG_YANG_2_SIMILARITY_2 = 'e'
>>> HUNG_YANG_2_SIMILARITY_3 = 'c'

>>> HUNG_YANG_3_SIMILARITY_1 = 'w1'
>>> HUNG_YANG_3_SIMILARITY_2 = 'w2'
>>> HUNG_YANG_3_SIMILARITY_3 = 'pk1'
>>> HUNG_YANG_3_SIMILARITY_4 = 'pk2'
>>> HUNG_YANG_3_SIMILARITY_5 = 'pk3'
>>> HUNG_YANG_3_SIMILARITY_6 = 'new1'
>>> HUNG_YANG_3_SIMILARITY_7 = 'new2'

>>> HUNG_YANG_4_SIMILARITY_1 = 'l'
>>> HUNG_YANG_4_SIMILARITY_2 = 'e'
>>> HUNG_YANG_4_SIMILARITY_3 = 'c'

>>> IANCU_SIMILARITY_1 = 1
>>> IANCU_SIMILARITY_2 = 2
>>> IANCU_SIMILARITY_3 = 3
>>> IANCU_SIMILARITY_4 = 4
>>> IANCU_SIMILARITY_5 = 5
>>> IANCU_SIMILARITY_6 = 6
>>> IANCU_SIMILARITY_7 = 7
>>> IANCU_SIMILARITY_8 = 8
>>> IANCU_SIMILARITY_9 = 9
>>> IANCU_SIMILARITY_10 = 10
>>> IANCU_SIMILARITY_11 = 11
>>> IANCU_SIMILARITY_12 = 12
>>> IANCU_SIMILARITY_13 = 13
>>> IANCU_SIMILARITY_14 = 14
>>> IANCU_SIMILARITY_15 = 15
>>> IANCU_SIMILARITY_16 = 16
>>> IANCU_SIMILARITY_17 = 17
>>> IANCU_SIMILARITY_18 = 18
>>> IANCU_SIMILARITY_19 = 19
>>> IANCU_SIMILARITY_20 = 20

>>> DENG_JIANG_FU_MONOTONIC_TYPE_1_1 = 1
>>> DENG_JIANG_FU_MONOTONIC_TYPE_1_2 = 2
>>> DENG_JIANG_FU_MONOTONIC_TYPE_1_3 = 3
>>> DENG_JIANG_FU_MONOTONIC_TYPE_1_4 = 4
>>> DENG_JIANG_FU_MONOTONIC_TYPE_2_1 = 5
>>> DENG_JIANG_FU_MONOTONIC_TYPE_2_2 = 6
>>> DENG_JIANG_FU_MONOTONIC_TYPE_2_3 = 7
>>> DENG_JIANG_FU_MONOTONIC_TYPE_2_4 = 8
>>> DENG_JIANG_FU_MONOTONIC_TYPE_3_1 = 9
>>> DENG_JIANG_FU_MONOTONIC_TYPE_3_2 = 10
>>> DENG_JIANG_FU_MONOTONIC_TYPE_3_3 = 11

"""

DISTANCE_HAMMING = 'Hamming'
DISTANCE_EUCLIDEAN = 'Euclidean'
DISTANCE_NORMALIZED_HAMMING = 'normalized Hamming'
DISTANCE_NORMALIZED_EUCLIDEAN = 'normalized Euclidean'

WANGXIN_DISTANCE_1 = 1
WANGXIN_DISTANCE_2 = 2

LIANG_SHI_SIMILARITY_1 = 'e'
LIANG_SHI_SIMILARITY_2 = 's'
LIANG_SHI_SIMILARITY_3 = 'h'

HUNG_YANG_1_SIMILARITY_1 = 'l'
HUNG_YANG_1_SIMILARITY_2 = 'e'
HUNG_YANG_1_SIMILARITY_3 = 'c'

HUNG_YANG_2_SIMILARITY_1 = 'l'
HUNG_YANG_2_SIMILARITY_2 = 'e'
HUNG_YANG_2_SIMILARITY_3 = 'c'

HUNG_YANG_3_SIMILARITY_1 = 'w1'
HUNG_YANG_3_SIMILARITY_2 = 'w2'
HUNG_YANG_3_SIMILARITY_3 = 'pk1'
HUNG_YANG_3_SIMILARITY_4 = 'pk2'
HUNG_YANG_3_SIMILARITY_5 = 'pk3'
HUNG_YANG_3_SIMILARITY_6 = 'new1'
HUNG_YANG_3_SIMILARITY_7 = 'new2'

HUNG_YANG_4_SIMILARITY_1 = 'l'
HUNG_YANG_4_SIMILARITY_2 = 'e'
HUNG_YANG_4_SIMILARITY_3 = 'c'

IANCU_SIMILARITY_1 = 1
IANCU_SIMILARITY_2 = 2
IANCU_SIMILARITY_3 = 3
IANCU_SIMILARITY_4 = 4
IANCU_SIMILARITY_5 = 5
IANCU_SIMILARITY_6 = 6
IANCU_SIMILARITY_7 = 7
IANCU_SIMILARITY_8 = 8
IANCU_SIMILARITY_9 = 9
IANCU_SIMILARITY_10 = 10
IANCU_SIMILARITY_11 = 11
IANCU_SIMILARITY_12 = 12
IANCU_SIMILARITY_13 = 13
IANCU_SIMILARITY_14 = 14
IANCU_SIMILARITY_15 = 15
IANCU_SIMILARITY_16 = 16
IANCU_SIMILARITY_17 = 17
IANCU_SIMILARITY_18 = 18
IANCU_SIMILARITY_19 = 19
IANCU_SIMILARITY_20 = 20

DENG_JIANG_FU_MONOTONIC_TYPE_1_1 = 1
DENG_JIANG_FU_MONOTONIC_TYPE_1_2 = 2
DENG_JIANG_FU_MONOTONIC_TYPE_1_3 = 3
DENG_JIANG_FU_MONOTONIC_TYPE_1_4 = 4
DENG_JIANG_FU_MONOTONIC_TYPE_2_1 = 5
DENG_JIANG_FU_MONOTONIC_TYPE_2_2 = 6
DENG_JIANG_FU_MONOTONIC_TYPE_2_3 = 7
DENG_JIANG_FU_MONOTONIC_TYPE_2_4 = 8
DENG_JIANG_FU_MONOTONIC_TYPE_3_1 = 9
DENG_JIANG_FU_MONOTONIC_TYPE_3_2 = 10
DENG_JIANG_FU_MONOTONIC_TYPE_3_3 = 11

from . import distances
from . import similarities
from . import misc
from . import sets

__all__ = (
    "distances",
    "similarities",
    "misc",
    "sets"
)



import numpy as np

from .utils._measure_input_validation import check_sets_cardinality
from .sets import FuzzySet


def fuzzy_divergence(A: FuzzySet, B: FuzzySet):
    """ Fuzzy Divergence proposed by J. Fan, W. Xie, from the related article: 
    "Distance measure and induced fuzzy entropy"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    ab_diff = A.membership_values - B.membership_values
    ba_diff = B.membership_values - A.membership_values

    return np.sum(
        2.0 -
        (1.0 - A.membership_values + B.non_membership_values) * np.exp(ab_diff) -
        (1.0 - B.membership_values + A.membership_values) * np.exp(ba_diff)
    )


def fuzzy_index(A: FuzzySet, coeff: int):
    """ Fuzzy Index  T. Chaira, A.R. Ray, from the related article: 
    "Threshold selection using fuzzy set theory"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    coeff: int
        Coefficient of the index.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    n = len(A)
    multiplier = 2.0 / float(np.power(n, 1.0 / coeff))

    return multiplier * np.sum(
        np.minimum(
            A.membership_values, 1.0 - A.membership_values
        ) ** coeff
    ) ** 1.0 / coeff





import numpy as np
from typing import Iterable, Tuple
from collections.abc import Callable
from tqdm import tqdm

from ..sets import FuzzySet, IntuitionisticFuzzySet
from .classifiers import classify, FuzzyTextClassifier
from .image_processing import threshold


def calculate_documents_membership(data: Iterable, membership_weight: float, non_membership_weight: float) -> Tuple[list, np.ndarray, np.ndarray]:
    """ Calculates the Fuzzy Set of each class from token counts.
    Proposed by P. Intarapaiboon from the related article: 
    "Text classification using similarity measures on intuitionistic fuzzy sets".

    Parameters
    ----------
    data : Iterable
        Token counts of a document dataset.
    membership_weight : float
        Weight used to calculate the membership values of each token's membership value.
    non_membership_weight : float
        Weight used to calculate the non-membership values of each token's membership value.
    
        
    Returns
    -------
    list[FuzzySet]
        List of IntuitionisticFuzzySets for each class.
    np.ndarray : means
        Mean values for each token in data.
    np.ndarray : stds
        Standard deviation values for each token in data.
    
    See also
    --------
    sklearn.feature_extraction.text.CountVectorizer
    """    
    means = np.mean(data, axis=0, dtype=np.float64)  # mean of each sample
    stds = np.std(data, axis=0, dtype=np.float64, ddof=1)  # std of each sample

    # calculate membership values of each word for each document
    z = (data - means) / stds
    np.nan_to_num(z, copy=False, nan=0)

    m = membership_weight / (1. + np.exp(-z))
    v = non_membership_weight / (1. + np.exp(z))
    p = 1 - m - v
    
    sets = [
        IntuitionisticFuzzySet(_m, v[i], p[i])
        for i, _m in enumerate(m)
    ]

    return sets, means, stds


def confidence_degree(predicted_class_distance: float, other_classes_distance: Iterable[float]) -> np.float64:
    """ Degree of Confidence proposd by A.G. Hatzimichailidis, G.A. Papakostas, V.G. Kaburlasos
    from the related article: 
    "A Novel Distance Measure of Intuitionistic Fuzzy Sets and Its Application to Pattern Recognition Problems"

    Parameters
    ----------
    predicted_class_distance : float
        Distance calculated between the sample and the predicted class.
    other_classes_distance: Iterable[float]
        An Iterable consisted of the distances calculated between the sample and every other class.
        
    Returns
    -------
    numpy.float64
        Degree of Confidence.
    """
    return np.sum(np.abs(predicted_class_distance - other_classes_distance))


def compactness(A: FuzzySet, shape : tuple) -> np.float64:
    """ Image geometry proposd by S.K. Pal, A. Rosenfeld, from the related article: 
    "Image enhancement and thresholding by optimization of fuzzy compactness"

    Parameters
    ----------
    A : FuzzySet
        Fuzzy set of an image, containing a 2-d np.ndarray as an image.
    shape : Tuple[float, float]
        Shape of the original image.
        
    Returns
    -------
    numpy.float64
        Compactness of the image.

    Raises
    ------
    ValueError
        In case the shape propvided is not a tuple of 2 values.
    """
    if len(shape) != 2:
        ValueError("Shape provided must have 2 elements (width and height), not {}.".format(len(shape)))
    
    shaped_memberships = A.membership_values.reshape(shape)
    
    a = np.sum(A.membership_values)
    p = np.sum(
        np.absolute(
            shaped_memberships[:, :-1] -
            shaped_memberships[:, 1:]
        ) +
        np.absolute(
            shaped_memberships[:, :-1] -
            shaped_memberships[1:, :]
        )
    )
    return a / (p ** 2)


def check_similarity_conditions(measure_caller: Callable, **measure_kwargs):
    """ Checks if measures satisfies the required conditions.

    Checks whether the measure provided satisfies the following conditions:

    .. math:: S(A, B) >= 0\ and\ S(A, B) <= 1
    .. math:: S(A, B) == S(B, A)
    .. math:: S(A, A) == 1

    The algorithm conducts two tests. The first one creates a list from `np.arange(start=0.0, stop=1.01, step=0.01)`
    and 10.000 creates two sets A and B with a random size (1-1000) of (non)membership values with random values picked from
    the range above, checking if the conditions apply.

    The second test creates 4 sets with random sizes (1-1000), with random values using `np.random.rand(random_size)`
     and checks for the conditions again. 
    The first two sets only contain membership and non-membership values, while the 2 last sets contain
    membership values and hesitation degrees. Checks if the condition applies when the measure
    is applied on all combination of the 4 sets.

    Parameters
    ----------
    measure_caller : Callable
        The measure to be tested.
    **measure_kwargs : measure arguments
        Passed to the measure_caller.

    Returns
    -------
    numpy.float64
        Compactness of the image.

    """
    value_range = np.arange(start=0.0, stop=1.01, step=0.01)
    for _ in tqdm(range(10000), desc="Test 1 progress:"):
        random_size = np.random.randint(1, 1000)
        A = FuzzySet(np.random.choice(value_range, random_size), np.random.choice(value_range, random_size))
        B = FuzzySet(np.random.choice(value_range, random_size), np.random.choice(value_range, random_size))
        assert measure_caller(A, B, **measure_kwargs) >= 0.0 and measure_caller(A, B, **measure_kwargs) <= 1.0

        assert measure_caller(A, B, **measure_kwargs) == measure_caller(B, A, **measure_kwargs)

        assert measure_caller(A, A, **measure_kwargs) == 1.0 and measure_caller(B, B, **measure_kwargs) == 1.0

    for _ in tqdm(range(10000), desc="Test 2 progress:"):
        random_size = np.random.randint(1, 1000)
        A = IntuitionisticFuzzySet(np.random.rand(random_size), np.random.rand(random_size))
        B = IntuitionisticFuzzySet(np.random.rand(random_size), np.random.rand(random_size))
        C = IntuitionisticFuzzySet(np.random.rand(random_size), None, np.random.rand(random_size))
        D = IntuitionisticFuzzySet(np.random.rand(random_size), None, np.random.rand(random_size))

        from itertools import combinations
        for a, b in combinations([A, B, C, D], 2):    
            assert measure_caller(a, b, **measure_kwargs) >= 0.0 and measure_caller(a, b, **measure_kwargs) <= 1.0

            assert measure_caller(a, b, **measure_kwargs) == measure_caller(b, a, **measure_kwargs)

            assert measure_caller(a, a, **measure_kwargs) == 1.0 and measure_caller(b, b, **measure_kwargs) == 1.0

    print("Similarity passes all tests!")


__all__ = (
    "classify",
    "FuzzyTextClassifier",
    "threshold",
    "calculate_documents_membership",
    "confidence_degree",
    "compactness",
    "check_similarity_conditions"
)


import numpy as np
from typing import Iterable
import warnings

from .utils._measure_input_validation import check_weights, check_p, check_sets_cardinality
from .sets import FuzzySet, IntuitionisticFuzzySet
from . import HUNG_YANG_1_SIMILARITY_1, HUNG_YANG_1_SIMILARITY_2, HUNG_YANG_1_SIMILARITY_3
from . import HUNG_YANG_2_SIMILARITY_1, HUNG_YANG_2_SIMILARITY_2, HUNG_YANG_2_SIMILARITY_3
from . import HUNG_YANG_3_SIMILARITY_1, HUNG_YANG_3_SIMILARITY_2, HUNG_YANG_3_SIMILARITY_3, HUNG_YANG_3_SIMILARITY_4, HUNG_YANG_3_SIMILARITY_5, HUNG_YANG_3_SIMILARITY_6, HUNG_YANG_3_SIMILARITY_7
from . import HUNG_YANG_4_SIMILARITY_1, HUNG_YANG_4_SIMILARITY_2, HUNG_YANG_4_SIMILARITY_3
from . import LIANG_SHI_SIMILARITY_1, LIANG_SHI_SIMILARITY_2, LIANG_SHI_SIMILARITY_3
from . import IANCU_SIMILARITY_1, IANCU_SIMILARITY_2, IANCU_SIMILARITY_3, IANCU_SIMILARITY_4, IANCU_SIMILARITY_5, IANCU_SIMILARITY_6, IANCU_SIMILARITY_7, IANCU_SIMILARITY_8, IANCU_SIMILARITY_9, IANCU_SIMILARITY_10, \
    IANCU_SIMILARITY_11, IANCU_SIMILARITY_12, IANCU_SIMILARITY_13, IANCU_SIMILARITY_14, IANCU_SIMILARITY_15, IANCU_SIMILARITY_16, IANCU_SIMILARITY_17, IANCU_SIMILARITY_18, IANCU_SIMILARITY_19, IANCU_SIMILARITY_20
from . import DENG_JIANG_FU_MONOTONIC_TYPE_1_1, DENG_JIANG_FU_MONOTONIC_TYPE_1_2, \
    DENG_JIANG_FU_MONOTONIC_TYPE_1_3, DENG_JIANG_FU_MONOTONIC_TYPE_1_4, DENG_JIANG_FU_MONOTONIC_TYPE_2_1, \
    DENG_JIANG_FU_MONOTONIC_TYPE_2_2, DENG_JIANG_FU_MONOTONIC_TYPE_2_3, DENG_JIANG_FU_MONOTONIC_TYPE_2_4, \
    DENG_JIANG_FU_MONOTONIC_TYPE_3_1, DENG_JIANG_FU_MONOTONIC_TYPE_3_2, DENG_JIANG_FU_MONOTONIC_TYPE_3_3


def dengfeng_chuntian(A: FuzzySet, B: FuzzySet, p: int = 1, weights: Iterable = None):
    """ Similarity proposed by L. Dengfeng and C. Chuntian, from the related article: 
    "New similarity measures of intuitionistic fuzzy sets and application to pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    p : int
        Positive integer >= 1.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p)
    check_weights(weights, len(A))

    fA = (A.membership_values + 1 - A.non_membership_values) / 2.0
    fB = (B.membership_values + 1 - B.non_membership_values) / 2.0
    n = len(A)
    if weights is None:
        return 1 - 1 / (n ** (1 / p)) * (np.sum(np.absolute(fA - fB) ** p) ** (1 / p))
    else:
        return 1 - np.power(np.sum(weights * np.power(np.absolute(fA - fB), p)), (1.0 / float(p)))


def liang_shi(A: FuzzySet, B: FuzzySet,  similarity_type: str = LIANG_SHI_SIMILARITY_1, p: int = 1, weights: Iterable = None , omegas: Iterable = [0.5, 0.3, 0.2]):
    """ Similarity proposed by Z. Liang and P. Shi, from the related article: 
    "Similarity measures on intuitionistic fuzzy sets""

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type : str, optional 
        Type of computed similarity:
        
        >>> LIANG_SHI_SIMILARITY_1
        >>> LIANG_SHI_SIMILARITY_2
        >>> LIANG_SHI_SIMILARITY_3

    p : int
        Positive integer >= 1.
    weights : list of floats
        List of weights for each membership/non-membership value.
    omegas: Iterable
        An iterable with 3 elements, with their sum equal to 1.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p)
    check_weights(weights, len(A), similarity_type, LIANG_SHI_SIMILARITY_3)

    if similarity_type == LIANG_SHI_SIMILARITY_3:
        if weights is None:
            weights = np.full(len(A), 1 / float(len(A)))

        omegas = np.array(omegas)
        if len(omegas.shape) == 1:
            omegas = np.expand_dims(omegas, axis=1)

        if not np.isclose(np.sum(omegas), 1.0):
            raise ValueError(
                "Sum of omegas parameter must be equal to 1, not {}".format(np.sum(omegas))
            )

    if similarity_type == LIANG_SHI_SIMILARITY_1:
        n = len(A)
        delta_memberships = np.absolute(A.membership_values - B.membership_values)
        delta_non_memberships = np.absolute(A.non_membership_values - B.non_membership_values)

        return 1.0 - 1.0 / np.power(n, 1.0 / p) * np.power(
                np.sum(
                    np.power((delta_memberships + delta_non_memberships) / 2.0, p)
                ),
                1.0 / p
            )
    elif similarity_type == LIANG_SHI_SIMILARITY_2:
        n = len(A)
        mA = (A.membership_values + 1.0 - A.non_membership_values) / 2.0
        mB = (B.membership_values + 1.0 - B.non_membership_values) / 2.0
        mA1 = (A.membership_values + mA) / 2.0
        mB1 = (B.membership_values + mB) / 2.0
        mA2 = (mA + 1.0 - A.non_membership_values) / 2.0
        mB2 = (mB + 1.0 - B.non_membership_values) / 2.0
        fS1 = np.absolute(mA1 - mB1) / 2.0
        fS2 = np.absolute(mA2 - mB2) / 2.0
        return 1.0 - (1.0 / np.power(n, 1.0 / p)) * np.power(
            np.sum(
                np.power(fS1 + fS2, p)
            ),
            1.0 / p
        )
    elif similarity_type == LIANG_SHI_SIMILARITY_3:
        delta_memberships = np.absolute(A.membership_values - B.membership_values)
        delta_non_memberships = np.absolute(A.non_membership_values - B.non_membership_values)

        f1 = (delta_memberships + delta_non_memberships) / 2.0
        mA = (A.membership_values + 1.0 - A.non_membership_values) / 2.0
        mB = (B.membership_values + 1.0 - B.non_membership_values) / 2.0
        f2 = np.absolute(mA - mB)
        iA = mA - A.membership_values
        iB = mB - B.membership_values
        f3 = np.maximum(iA, iB) - np.minimum(iA, iB)
        fs = np.array([f1, f2, f3])
        return 1.0 - np.power(
            np.sum(
                weights * np.power(np.sum(omegas * fs), p)
            ),
            1.0 / p
        )
    else:
        raise ValueError(
            "similarity_type parameter must be LIANG_SHI_SIMILARITY_1, LIANG_SHI_SIMILARITY_2, LIANG_SHI_SIMILARITY_3 or LIANG_SHI_SIMILARITY_4.")


def park_kwun_lim(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet, p: int = 1, weights: Iterable = None):
    """ Similarity proposed by A.H. Park, A.S. Park, Y.C. Kwun and K.M. Lim, from the related article:
    "New Similarity Measures on Intuitionistic Fuzzy Sets".

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    p : int
        Positive integer >= 1.
    weights : list of floats
        List of weights for each membership/non-membership value.
    
        
    Returns
    -------
    numpy.float64
    The similarity between the two sets provided.

    """
    check_sets_cardinality(A, B)
    check_p(p)

    n = len(A)
    delta_memberships = np.abs(A.membership_values - B.membership_values)
    delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
    delta_hesitations = np.abs(A.hesitation_degrees - B.hesitation_degrees)
    if weights is None:
        return 1.0 - (1.0 / np.power(n, 1.0 / p)) * np.power(
            np.sum(
                np.power(delta_memberships + delta_non_memberships + delta_hesitations, p)
            ) / 2.0,
            1.0 / p
        )
    else:
        check_weights(weights, len(A))
        return 1.0 - np.power(
            np.sum(
                weights * np.power(delta_memberships + delta_non_memberships + delta_hesitations, p)
            ) / 2.0,
            1.0 / p
        )


def mitchell(A: FuzzySet, B: FuzzySet, p: int = 1, weights: Iterable = None):
    """ Similarity proposed by H.B. Mitchell, from the related article: 
    "On the Dengfeng-Chuntian similarity measure and its application to pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    p : int
        Positive integer >= 1.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p)
    check_weights(weights, len(A))

    if weights is None:
        weights = np.full(len(A), 1.0 / float(len(A)))
    
    D_m = 1.0 - np.power(
        np.sum(
            weights * 
            np.power(np.absolute(A.membership_values - B.membership_values), p)
        ), (1.0 / p)
    )
    D_f = 1.0 - np.power(
        np.sum(
            weights * 
            np.power(np.absolute(A.non_membership_values - B.non_membership_values), p)
        ), (1.0 / p)
    )

    return (D_m + D_f) / 2.0


def julian_hung_lin(A: FuzzySet, B: FuzzySet, p: int = 1, weights: Iterable = None):
    """ Similarity proposed by P. Julian, K.C. Hung and S.J. Lin, from the related article: 
    "On the Mitchell similarity measure and its application to pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    p : int
        Positive integer >= 1.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p)
    check_weights(weights, len(A))

    if weights is None:
        weights = np.full(len(A), 1 / float(len(A)))
    
    fA = 1.0 - A.non_membership_values
    fB = 1.0 - B.non_membership_values
    return 1.0 - np.power(
        np.sum(
            weights * 
            np.power(np.absolute(A.membership_values - B.membership_values), p)
        ), 1.0 / p
    ) + np.power(
        np.sum(
            weights * 
            np.power(np.absolute(fA - fB), p)
        ), 1.0 / p
    )


def hung_yang_1(A: FuzzySet, B: FuzzySet, similarity_type: str = HUNG_YANG_1_SIMILARITY_1, weights: Iterable = None):
    """ Similarity proposed by W.L. Hung and M.S. Yang, from the related article: 
    "Similarity measures of intuitionistic fuzzy sets based on Hausdorff similarity"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type: str, optional
        Type of computed similarity:
        
        >>> HUNG_YANG_1_SIMILARITY_1
        >>> HUNG_YANG_1_SIMILARITY_2
        >>> HUNG_YANG_1_SIMILARITY_3
    
    weights: List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)

    if weights is not None:
        weights = np.array(weights)

    if weights is None:
        D = np.sum(np.maximum(np.absolute(A.membership_values - B.membership_values),
                   np.absolute(A.non_membership_values - B.non_membership_values))) / float(len(A))
    else:
        check_weights(weights, len(A))
        D = np.sum(weights * np.maximum(np.absolute(A.membership_values - B.membership_values),
                   np.absolute(A.non_membership_values - B.non_membership_values)))

    if similarity_type == HUNG_YANG_1_SIMILARITY_1:
        return 1 - D
    elif similarity_type == HUNG_YANG_1_SIMILARITY_2:
        tp = np.exp(-1)
        return (np.exp(-D) - tp) / float(1 - tp)
    elif similarity_type == HUNG_YANG_1_SIMILARITY_3:
        return (1 - D) / float(1 + D)
    else:
        raise ValueError(
            "similarity_type parameter must be HUNG_YANG_1_SIMILARITY_1, HUNG_YANG_1_SIMILARITY_2 or HUNG_YANG_1_SIMILARITY_3.")


def ye(A: FuzzySet, B: FuzzySet, weights: Iterable = None):
    """ Similarity proposed by J. Ye, from the related article: 
    "Cosine similarity measures for intuitionistic fuzzy sets and their applications"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)

    n = len(A)
    nominator = (
        A.membership_values * B.membership_values + 
        A.non_membership_values * B.non_membership_values
    )
    denominator = (
        np.sqrt(A.membership_values ** 2.0 + A.non_membership_values ** 2.0) *
        np.sqrt(B.membership_values ** 2.0 + B.non_membership_values ** 2.0)
    )

    if weights is None:
        return 1.0 / n * np.sum(nominator / denominator)
    else:
        check_weights(weights, len(A))
        return sum(weights * nominator / denominator)


def hwang_yang(A: FuzzySet, B: FuzzySet):
    """ Similarity proposed by C.M. Hwang and M.S. Yang, from the related article: 
    "Modified cosine similarity measure between intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)

    n = len(A)
    
    C1 = ye(A, B)

    fA = 1.0 + A.membership_values - A.non_membership_values
    fB = 1.0 + B.membership_values - B.non_membership_values
    
    nominator = fA * fB + (4.0 * A.non_membership_values * B.non_membership_values)
    
    denominator = (
        np.sqrt(fA ** 2.0 + (2.0 * A.non_membership_values) ** 2.0) *
        np.sqrt(fB ** 2.0 + (2.0 * B.non_membership_values) ** 2.0)
    )
    C2 = 1.0 / n * np.sum(nominator / denominator)

    nominator = (
        ((1.0 - A.membership_values) * (1.0 - B.membership_values)) + 
        ((1.0 - A.non_membership_values) * (1.0 - B.non_membership_values))
    )
    denominator = (
        np.sqrt((1.0 - A.membership_values) ** 2.0 + (1.0 - B.membership_values) ** 2.0) *
        np.sqrt((1.0 - A.non_membership_values) ** 2.0 + (1.0 - B.non_membership_values) ** 2.0)
    )
    C3 = 1.0 / n * np.sum(nominator / denominator)

    return (C1 + C2 + C3) / 3.0


def hung_yang_2(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet, similarity_type: str = HUNG_YANG_2_SIMILARITY_1, a: int = 1):
    """ Similarity proposed by W.L. Hung and M.S. Yang, from the related article: 
    "On the J-divergence of intuitionistic fuzzy sets with its applications to pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type : str, optional
        Type of computed similarity:

        >>> HUNG_YANG_2_SIMILARITY_1
        >>> HUNG_YANG_2_SIMILARITY_2
        >>> HUNG_YANG_2_SIMILARITY_3

    a: case of divergence measure. Positive integer >= 1.
    
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    Raises:
        ValueError if a is < 1.
    """
    def _log(arr):
        invalids = np.logical_or(
            arr <= 0, 
            np.isnan(arr), 
            np.isinf(arr)
        )
        invalid_indxs = np.where(invalids)
        valid_indxs = np.where(np.logical_not(invalids))
        arr[invalid_indxs] = 0.0
        arr[valid_indxs] = np.log(arr[valid_indxs])
        return arr
    
    check_sets_cardinality(A, B)
    
    n = len(A)
    if a < 1 and not isinstance(a, int):
        raise ValueError("a parameter must be an integer >= 1.")

    if a == 1:
        U_a = np.log(2.0)

        LABm = (
            (A.membership_values + B.membership_values) * _log((A.membership_values + B.membership_values) / 2) -
            A.membership_values * _log(A.membership_values) - 
            B.membership_values * _log(B.membership_values)
        )

        LABv = (
            (A.non_membership_values + B.non_membership_values) * _log((A.non_membership_values + B.non_membership_values) / 2) - 
            A.non_membership_values *  _log(A.non_membership_values) - 
            B.non_membership_values * _log(B.non_membership_values)
        )

        LABp = (
            (A.hesitation_degrees + B.hesitation_degrees) * _log((A.hesitation_degrees + B.hesitation_degrees) / 2) -
            A.hesitation_degrees * _log(A.hesitation_degrees) - 
            B.hesitation_degrees * _log(B.hesitation_degrees)
        )
        ja = -0.5 * (LABm + LABv + LABp)
    else:
        U_a = 1.0 / (float(a) - 1.0) * (1 - 1 / float(2.0 ** (a - 1.0)))
        TmAB = (
                np.power((A.membership_values + B.membership_values) / 2.0, a) - 
            0.5 * (
                np.power(A.membership_values, a) + 
                np.power(B.membership_values, a)
            )
        )
        TvAB = (
            np.power((A.non_membership_values + B.non_membership_values) / 2.0, a) - 
            0.5 * (
                np.power(A.non_membership_values, a) +
                np.power(B.non_membership_values, a)
            )
        )
        TpAB = (
            np.power((A.hesitation_degrees + B.hesitation_degrees) / 2.0, a) - 
            0.5 * (
                np.power(A.hesitation_degrees, a) +
                np.power(B.hesitation_degrees, a)
            )
        )
        ja = -1 / (a - 1) * (TmAB + TvAB + TpAB)

    J_a = np.sum(ja) / float(n)

    if (similarity_type == HUNG_YANG_2_SIMILARITY_1):
        return (U_a - J_a) / U_a
    elif (similarity_type == HUNG_YANG_2_SIMILARITY_2):
        return (np.exp(-J_a) - np.exp(-U_a)) / float(1 - np.exp(-U_a))
    elif (similarity_type == HUNG_YANG_2_SIMILARITY_3):
        return (U_a - J_a) / np.array((1 + J_a) * U_a)
    else:
        raise ValueError(
            "similarity_type parameter must be HUNG_YANG_2_SIMILARITY_1, HUNG_YANG_2_SIMILARITY_2 or HUNG_YANG_2_SIMILARITY_3.")


def zhang_fu(A: FuzzySet, B: FuzzySet):
    """ Similarity proposed by C. Zhang and H. Fu, from the related article: 
    "Similarity measures on three kinds of fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    
    n = len(A)
    dA = A.membership_values + (1.0 - A.membership_values - A.non_membership_values) * A.membership_values
    aA = A.non_membership_values + (1.0 - A.membership_values - A.non_membership_values) * A.non_membership_values

    dB = B.membership_values + (1.0 - B.membership_values - B.non_membership_values) * B.membership_values
    aB = B.non_membership_values + (1.0 - B.membership_values - B.non_membership_values) * B.non_membership_values
    
    return 1.0 - (1.0 / (2.0 * n)) * np.sum(np.abs(dA - dB) + np.abs(aA - aB))


def hung_yang_3(A: FuzzySet, B: FuzzySet, similarity_type: str = HUNG_YANG_3_SIMILARITY_1):
    """ Similarity proposed by W.L. Hung and M.S. Yang, from the related article: 
    "On similarity measures between intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type : str, optional
        Type of computed similarity:

        >>> HUNG_YANG_3_SIMILARITY_1
        >>> HUNG_YANG_3_SIMILARITY_2
        >>> HUNG_YANG_3_SIMILARITY_3
        >>> HUNG_YANG_3_SIMILARITY_4
        >>> HUNG_YANG_3_SIMILARITY_5
        >>> HUNG_YANG_3_SIMILARITY_6
        >>> HUNG_YANG_3_SIMILARITY_7
        
    Returns
    -------
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)

    n = float(len(A))

    if similarity_type == HUNG_YANG_3_SIMILARITY_1:
        num = np.minimum(A.membership_values, B.membership_values) + \
            np.minimum(A.non_membership_values, B.non_membership_values)
        denom = np.maximum(A.membership_values, B.membership_values) + \
            np.maximum(A.non_membership_values, B.non_membership_values)
        D = np.sum(np.divide(num, denom.astype(float)))
        return D / n
    elif similarity_type == HUNG_YANG_3_SIMILARITY_2:
        ptr2 = 0.5 * np.absolute(A.membership_values - B.membership_values) + \
            np.absolute(A.non_membership_values - B.non_membership_values)
        D = np.sum(np.subtract(1, ptr2))
        return D / n
    elif similarity_type == HUNG_YANG_3_SIMILARITY_3:
        num = np.minimum(A.membership_values, B.membership_values) + \
            np.minimum(A.non_membership_values, B.non_membership_values)
        denom = np.maximum(A.membership_values, B.membership_values) + \
            np.maximum(A.non_membership_values, B.non_membership_values)
        D1 = np.sum(num)
        D2 = np.sum(denom)
        return D1 / float(D2)
    elif similarity_type == HUNG_YANG_3_SIMILARITY_4:
        dif_m = np.absolute(A.membership_values - B.membership_values)
        dif_v = np.absolute(A.non_membership_values - B.non_membership_values)
        return 1 - 0.5 * (dif_m.max() + dif_v.max())
    elif similarity_type == HUNG_YANG_3_SIMILARITY_5:
        num = np.absolute(A.membership_values - B.membership_values) + \
            np.absolute(A.non_membership_values - B.non_membership_values)
        denom = np.absolute(A.membership_values + B.membership_values) + \
            np.absolute(A.non_membership_values + B.non_membership_values)
        D1 = np.sum(num)
        D2 = np.sum(denom)
        return 1 - (D1 / float(D2))
    elif similarity_type == HUNG_YANG_3_SIMILARITY_6:
        num = 1 - np.exp(-0.5 * sum(np.absolute(A.membership_values - B.membership_values) +
                         np.absolute(A.non_membership_values - B.non_membership_values)))
        denum = 1 - np.exp(-n)
        return 1 - num / denum
    elif similarity_type == HUNG_YANG_3_SIMILARITY_7:
        num = 1 - np.exp(
            -0.5 * np.sum(np.absolute(np.sqrt(A.membership_values) - np.sqrt(B.membership_values)) + np.absolute(np.sqrt(A.non_membership_values) - np.sqrt(B.non_membership_values))))
        denum = 1 - np.exp(-n)
        return 1 - num / denum
    else:
        raise ValueError(
            "similarity_type parameter must be HUNG_YANG_3_SIMILARITY_1, HUNG_YANG_3_SIMILARITY_2, HUNG_YANG_3_SIMILARITY_3, HUNG_YANG_3_SIMILARITY_4, HUNG_YANG_3_SIMILARITY_,5, HUNG_YANG_3_SIMILARITY_6 or HUNG_YANG_3_SIMILARITY_7.")


def chen_1(A: FuzzySet, B: FuzzySet, weights: Iterable = None):
    """ Similarity proposed by S.M. Chen, from the related article: 
    "Measures of similarity between vague sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_weights(weights, len(A))

    if weights is None:
        weights = np.full(len(A), 1.0 / len(A))

    sA = A.membership_values - A.non_membership_values
    sB = B.membership_values - B.non_membership_values
    
    return np.sum(
        weights * (
            1.0 - np.absolute((sA - sB) / 2.0)
        )
    ) / np.sum(weights)


def hung_yang_4(A: FuzzySet, B: FuzzySet, similarity_type: str = HUNG_YANG_4_SIMILARITY_1, p: int = 1):
    """ Similarity proposed by W.L. Hung and M.S. Yang, from the related article: 
    "Similarity measures of intuitionistic fuzzy sets based on Lp metric"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type : str, optional
        Type of computed similarity:

        >>> HUNG_YANG_4_SIMILARITY_1
        >>> HUNG_YANG_4_SIMILARITY_2 
        >>> HUNG_YANG_4_SIMILARITY_3

    p : int
        Positive integer >= 1.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p)
    
    D = np.sum(np.power((np.absolute(np.power(A.membership_values - B.membership_values, p)) +
               np.power(np.absolute(A.non_membership_values - B.non_membership_values), p)), (1.0 / p)))

    D = D / float(len(A))
    f1 = 2 ** (1.0 / float(p))

    if similarity_type == HUNG_YANG_4_SIMILARITY_1:
        return (f1 - D) / float(f1)
    elif similarity_type == HUNG_YANG_4_SIMILARITY_2:
        return (np.exp(-D) - np.exp(-f1)) / float(1 - np.exp(-f1))
    elif similarity_type == HUNG_YANG_4_SIMILARITY_3:
        return (f1 - D) / float(f1 * (1.0 + D))
    else:
        raise ValueError(
            "similarity_type parameter must be HUNG_YANG_4_SIMILARITY_1, HUNG_YANG_4_SIMILARITY_2 or HUNG_YANG_4_SIMILARITY_3.")


def hong_kim(A: FuzzySet, B: FuzzySet, weights: Iterable = None, a: int = 1, b: int = 0, c: int = 0):
    """ Similarity proposed by D.H. Hong and C.Kim, from the related article: 
    "A note on similarity measures between vague sets and between elements"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
    a, b, c: int
        Must satisfy the condition: a >= c >= 0 >= b.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_weights(weights, len(A))
    n = len(A)
    
    if not (a >= c >= 0 >= b):
        raise ValueError(
            "Parameters a, b and c must satisfy the condition a >= c >= 0 >= b, got {} {} {}".format(a, b, c))

    if weights is None:
        D = np.sum(
            1 - (
                np.abs(A.membership_values - B.membership_values) +
                np.abs(A.non_membership_values - B.non_membership_values)
            ) / 2.0
        )
        return D / float(n)
    else:
        denom = float(a + b + c)

        num = (
            (a * np.absolute(A.membership_values - B.membership_values)) +
            (b * np.absolute(A.non_membership_values - B.non_membership_values)) +
            (c * np.absolute(B.membership_values + 
                             B.non_membership_values - 
                             A.membership_values - 
                             A.non_membership_values))
        )

        D = np.sum(weights * (1 - num / denom))

        return D


def chen_2(A: FuzzySet, B: FuzzySet, weights: Iterable = None, a: int = 1, b: int = 0, c: int = 0) -> float:
    """ Similarity proposed by S.M. Chen, from the related article: 
        "Similarity measure between vague sets and between elements"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
    a, b, c: int
        Must satisfy the condition: a >= c >= 0 >= b.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_weights(weights, len(A))

    if not (a >= c >= 0 >= b):
        raise ValueError(
            "a, b and c parameters must meet the condition: a >= c >= 0 >= b.")

    if weights is None:
        weights = np.full(len(A), 1 / float(len(A)))

    denom = a - b
    num = (a * (A.membership_values - B.membership_values)) + (b * (A.non_membership_values - B.non_membership_values)
                                                               ) - (c * (A.membership_values - B.membership_values + A.non_membership_values - B.non_membership_values))
    D = np.sum(weights * (1 - np.absolute(num) / float(denom)))
    D = np.array(D)

    return D / float(np.sum(weights))


def liu(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet, p: int = 1, weights: Iterable = None, a: float = 0.4, b: float = 0.3, c: float = 0.3):
    """ Similarity proposed by H.W. Liu, from the related article: 
    "New similarity measures between intuitionistic fuzzy sets and between elements"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    p : int
        Positive integer >= 1.
    weights : list of floats
        List of weights for each membership/non-membership value.
    a, b, c: float
        Sum of those parameters must be equal to 1.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p)
    check_weights(weights, len(A))
    n = len(A)

    if weights is None:
        D = np.sum((np.power(np.absolute(A.membership_values - B.membership_values), p)) + (np.power(np.absolute(A.non_membership_values - B.non_membership_values), p)) + (
            np.power(np.absolute(A.hesitation_degrees - B.hesitation_degrees), p)))
        return 1 - (D / float(2.0 * n)) ** (1.0 / p)
    else:
        prts = (a * np.power(np.absolute(A.membership_values - B.membership_values), p)) + (b * np.power(np.absolute(A.non_membership_values - B.non_membership_values), p)) + (
            c * np.power(np.absolute(A.hesitation_degrees - B.hesitation_degrees), p))
        D = np.sum(weights * prts)
        return 1 - D ** (1.0 / p)


def iancu(A: FuzzySet, B: FuzzySet, similarity_type: int = IANCU_SIMILARITY_1, lamda: int = 1):
    """ Similarities proposed by I. Iancu, from the related article: 
    "Intuitionistic fuzzy similarity measures based on Frank t-norms family"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type : int, optional
        Type of computed similarity:

        >>> IANCU_SIMILARITY_1 IANCU_SIMILARITY_2 ..., IANCU_SIMILARITY_20
    
    lambda: float
        Frank family of t-operator parameter. Different cases of input:
        0, 1, Inf and other. Used in all similarity_type cases *except*
        
        >>> IANCU_SIMILARITY_1, IANCU_SIMILARITY_2,  IANCU_SIMILARITY_3 
        >>> IANCU_SIMILARITY_4, IANCU_SIMILARITY_18
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    def tOperator(a: Iterable, b: Iterable, lamda: float):
        if (lamda == np.inf):
            return np.maximum(0, a + b - 1)
        elif (lamda == 0):
            return np.minimum(a, b)
        elif (lamda == 1):
            return a * b
        else:
            return np.log(1.0 + (
                (
                    (np.power(lamda, 1 - a) - 1.0) * (np.power(lamda, 1.0 - b) - 1.0)
                ) / (lamda - 1.0)
            ))
    delta_A = A.membership_values - A.non_membership_values
    delta_B = B.membership_values - B.non_membership_values
    n = len(A)
    if similarity_type == IANCU_SIMILARITY_1:  # S11
        return (
            (n + np.minimum(np.sum(delta_A), np.sum(delta_B))) /
            (n + np.maximum(np.sum(delta_A), np.sum(delta_B)))
        )
    elif similarity_type == IANCU_SIMILARITY_2:  # S11c
        nominator = n - np.maximum(np.sum(delta_A), np.sum(delta_B))
        denom = n - np.minimum(np.sum(delta_A), np.sum(delta_B))
        return nominator / np.array(denom).astype("float")
    elif similarity_type == IANCU_SIMILARITY_3:  # S19
        nominator = n + np.minimum(np.sum(delta_A),
                                   np.sum(delta_B))
        return nominator / (2.0 * n)
    elif similarity_type == IANCU_SIMILARITY_4:  # S19c
        nominator = n - np.maximum(np.sum(delta_A),
                                   np.sum(delta_B))
        return nominator / (2.0 * n)
    elif similarity_type == IANCU_SIMILARITY_5:  # S1
        return (
            (
                n + np.sum(
                    tOperator(A.membership_values, B.membership_values, lamda) +
                    tOperator(A.non_membership_values, B.non_membership_values, lamda) -
                    A.non_membership_values - B.non_membership_values
                    )
             ) / 
             (
                 n + np.max(np.sum(delta_A), np.sum(delta_B))
             )
        )
    elif similarity_type == IANCU_SIMILARITY_6:  # S1c
        nom = n + np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(
            A.non_membership_values, B.non_membership_values, lamda) - A.membership_values - B.membership_values)
        denom = n - np.minimum(np.sum(delta_A),
                               np.sum(delta_B))
        return nom / np.array(denom).astype("float")
    elif similarity_type == IANCU_SIMILARITY_7:  # S5
        nom = n + np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(
            A.non_membership_values, B.non_membership_values, lamda) - A.non_membership_values - B.non_membership_values)
        denom = n + np.sum(A.membership_values + B.membership_values - tOperator(A.membership_values,
                           B.membership_values, lamda) - tOperator(A.non_membership_values, B.non_membership_values, lamda))
        return nom / np.array(denom).astype("float")
    elif similarity_type == IANCU_SIMILARITY_8:  # S5c
        nom = n + np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(
            A.non_membership_values, B.non_membership_values, lamda) - A.membership_values - B.membership_values)
        denom = n + np.sum(A.non_membership_values + B.non_membership_values - tOperator(A.membership_values,
                           B.membership_values, lamda) - tOperator(A.non_membership_values, B.non_membership_values, lamda))
        return nom / np.array(np.sum(denom)).astype("float")
    elif similarity_type == IANCU_SIMILARITY_9:  # S14
        nom = n + np.minimum(np.sum(delta_A),
                             np.sum(delta_B))
        denom = n + np.sum(A.membership_values + B.membership_values - tOperator(A.membership_values,
                                                                                 B.membership_values, lamda) - tOperator(A.non_membership_values, B.non_membership_values, lamda))
        return nom / float(denom)
    elif similarity_type == IANCU_SIMILARITY_10:  # S14c
        nom = n - np.maximum(np.sum(delta_A),
                             np.sum(delta_B))
        denom = n + np.sum(A.non_membership_values + B.non_membership_values - tOperator(A.membership_values,
                                                                                         B.membership_values, lamda) - tOperator(A.non_membership_values, B.non_membership_values, lamda))
        return nom / float(denom)
    elif similarity_type == IANCU_SIMILARITY_11:  # S18
        nom = n + np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(
            A.non_membership_values, B.non_membership_values, lamda) - A.non_membership_values - B.non_membership_values)
        denom = 2.0 * n
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_12:  # S18c
        nom = n - np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(
            A.non_membership_values, B.non_membership_values, lamda) - A.membership_values - B.membership_values)
        denom = 2.0 * n
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_13:  # S2
        nom = 2.0 * n + np.sum(
            2.0 * tOperator(A.membership_values, B.membership_values, lamda) + 2.0 * tOperator(A.non_membership_values, B.non_membership_values, lamda) - A.membership_values - B.membership_values - A.non_membership_values - B.non_membership_values)
        denom = 2.0 * n + np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(A.non_membership_values, B.non_membership_values, lamda)) - np.minimum(
            np.sum(A.membership_values + B.non_membership_values), np.sum(B.membership_values + A.non_membership_values))
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_14:  # S6
        nom = 2.0 * n + np.sum(
            2.0 * tOperator(A.membership_values, B.membership_values, lamda) + 2.0 * tOperator(A.non_membership_values, B.non_membership_values, lamda) - A.membership_values - B.membership_values - A.non_membership_values - B.non_membership_values)
        denom = 2.0 * n
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_15:  # S13
        nom = np.minimum(np.sum(A.membership_values + B.non_membership_values), np.sum(B.membership_values + A.non_membership_values)) - np.sum(
            tOperator(A.membership_values, B.membership_values, lamda) + tOperator(A.non_membership_values, B.non_membership_values, lamda))
        denom = np.sum(A.membership_values + B.membership_values + A.non_membership_values + B.non_membership_values - 2.0 * tOperator(
            A.membership_values, B.membership_values, lamda) - 2 * tOperator(A.non_membership_values, B.non_membership_values, lamda))
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_16:  # S15
        nom = 2.0 * n + np.sum(tOperator(A.membership_values, B.membership_values, lamda) + tOperator(A.non_membership_values, B.non_membership_values, lamda)) - np.maximum(np.sum(A.membership_values + B.non_membership_values),
                                                                                                                                                                             np.sum(B.membership_values + A.non_membership_values))
        denom = 2.0 * n
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_17:  # S2"
        nom = np.sum(A.membership_values + B.membership_values + A.non_membership_values + B.non_membership_values - 2.0 * tOperator(
            A.membership_values, B.non_membership_values, lamda) - 2.0 * tOperator(B.membership_values, A.non_membership_values, lamda))
        denom = n - np.sum(tOperator(A.membership_values, B.non_membership_values, lamda) + tOperator(B.membership_values, A.non_membership_values, lamda)) + np.maximum(np.sum(A.membership_values + B.non_membership_values),
                                                                                                                                                                         np.sum(B.membership_values + A.non_membership_values))
        return nom / denom.astype(float)
    elif similarity_type == IANCU_SIMILARITY_18:  # S6"
        nom = np.sum(A.membership_values + B.membership_values + A.non_membership_values + B.non_membership_values - 2.0 * tOperator(
            A.membership_values, B.non_membership_values, lamda) - 2.0 * tOperator(B.membership_values, A.non_membership_values, lamda))
        denom = 2.0 * n
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_19:  # S13"
        nom = n + np.sum(tOperator(A.membership_values, B.non_membership_values, lamda) + tOperator(B.membership_values, A.non_membership_values, lamda)) - np.maximum(np.sum(A.membership_values + B.non_membership_values),
                                                                                                                                                                       np.sum(B.membership_values + A.non_membership_values))
        denom = 2.0 * n + np.sum(
            2.0 * tOperator(A.membership_values, B.non_membership_values, lamda) + 2.0 * tOperator(B.membership_values, A.non_membership_values, lamda) - A.membership_values - B.membership_values - A.non_membership_values - B.non_membership_values)
        return nom / denom
    elif similarity_type == IANCU_SIMILARITY_20:  # S15"
        nom = n + np.minimum(np.sum(A.membership_values + B.non_membership_values), np.sum(B.membership_values + A.non_membership_values)) - np.sum(
            tOperator(A.membership_values, B.non_membership_values, lamda) + tOperator(B.membership_values, A.non_membership_values, lamda))
        denom = 2.0 * n
        return nom / denom
    else:
        raise ValueError(
            "similarity_type parameter must be IANCU_SIMILARITY_1, IANCU_SIMILARITY_2, IANCU_SIMILARITY_3, ..., or IANCU_SIMILARITY_20.")


def song_wang_lei_xue(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet, weights: Iterable = None):
    """ Similarity proposed by Y. Song, X. Wang, L. Lei, A. Xue, from the related article: 
    "A novel similarity measure on intuitionistic fuzzy sets with its applications"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    n = len(A)

    if weights is None:
        ff = (1.0 / (2.0 * n)) * np.sum(
            np.sqrt(A.membership_values * B.membership_values) + (2 * np.sqrt(A.non_membership_values * B.non_membership_values)) + np.sqrt(A.hesitation_degrees * B.hesitation_degrees) + np.sqrt((1 - A.non_membership_values) * (1 - B.non_membership_values)))
        return ff
    else:
        check_weights(weights, len(A))
        return 1.0 / 2.0 * np.sum(
            weights * (np.sqrt(A.membership_values * B.membership_values) + 2.0 * np.sqrt(A.non_membership_values * B.non_membership_values) + np.sqrt(A.hesitation_degrees * B.hesitation_degrees) + np.sqrt((1 - A.non_membership_values) * (1 - B.non_membership_values))))


def intarapaiboon(A: FuzzySet, B: FuzzySet):
    """ Similarity proposed by P. Intarapaiboon, from the related article: 
    "A hierarchy-based similarity measure for intuitionistic fuzzy sets"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    n = len(A)
    delta_memberships = np.abs(A.membership_values - B.membership_values)
    delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
    return 1 - (1.0 / (2.0 * n) * np.sum(delta_memberships + delta_non_memberships))


def deng_jiang_fu(A: FuzzySet, B: FuzzySet, similarity_type: int = DENG_JIANG_FU_MONOTONIC_TYPE_1_1, p: int = None, u:float = None, v: float = None):
    """ Similarity proposed by G. Deng, Y. Jiang, J. Fu, from the related article: 
    "Monotonic similarity measures between intuitionistic fuzzy sets and their relationship with entropy and inclusion measure"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    similarity_type : int, optional
        Type of computed similarity:

        >>> DENG_JIANG_FU_MONOTONIC_TYPE_1_1
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_1_2
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_1_3
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_1_4
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_2_1
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_2_2
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_2_3
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_2_4
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_3_1
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_3_2
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_3_3

    p: float 
        must be >= 1. Used in all types *except* 
        
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_1_3 DENG_JIANG_FU_MONOTONIC_TYPE_2_3 DENG_JIANG_FU_MONOTONIC_TYPE_3_1 DENG_JIANG_FU_MONOTONIC_TYPE_3_2 DENG_JIANG_FU_MONOTONIC_TYPE_3_3
    u: float
        Must be positive. Used only in 
        
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_3_2
    v: float
        Must be positive. Used only in
        
        >>> DENG_JIANG_FU_MONOTONIC_TYPE_3_2
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_p(p, similarity_type, DENG_JIANG_FU_MONOTONIC_TYPE_1_3, DENG_JIANG_FU_MONOTONIC_TYPE_2_3, DENG_JIANG_FU_MONOTONIC_TYPE_3_1, DENG_JIANG_FU_MONOTONIC_TYPE_3_2, DENG_JIANG_FU_MONOTONIC_TYPE_3_3)

    if p is not None and similarity_type not in [DENG_JIANG_FU_MONOTONIC_TYPE_1_3, DENG_JIANG_FU_MONOTONIC_TYPE_2_3, DENG_JIANG_FU_MONOTONIC_TYPE_3_1, DENG_JIANG_FU_MONOTONIC_TYPE_3_2, DENG_JIANG_FU_MONOTONIC_TYPE_3_3]:
        warnings.warn("Ignoring parameter p (not used in provided similarity_type).")
    if u is not None and similarity_type != DENG_JIANG_FU_MONOTONIC_TYPE_3_2:
        warnings.warn("Ignoring parameter u (not used in provided similarity_type).")
    if v is not None and similarity_type != DENG_JIANG_FU_MONOTONIC_TYPE_3_2:
        warnings.warn("Ignoring parameter v (not used in provided similarity_type).")
    
    if similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_1_1:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        min_memberships = np.minimum(A.membership_values, B.membership_values)
        min_non_memberships = np.minimum(A.non_membership_values, B.non_membership_values)
        max_memberships = np.maximum(A.membership_values, B.membership_values)
        max_non_memberships = np.maximum(A.non_membership_values, B.non_membership_values)
        return 1.0 / n * np.sum(
            (min_memberships + min_non_memberships) / 
            (max_memberships + max_non_memberships + 
            delta_memberships + delta_non_memberships)
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_1_2:
        n = len(A)
        min_memberships = np.minimum(A.membership_values, B.membership_values)
        min_non_memberships = np.minimum(A.non_membership_values, B.non_membership_values)
        max_memberships = np.maximum(A.membership_values, B.membership_values)
        max_non_memberships = np.maximum(A.non_membership_values, B.non_membership_values)
        return (
            np.sum(
                (min_memberships + min_non_memberships) /
                (2.0 * (max_memberships + max_non_memberships))
            ) /
            (
                n - np.sum(
                    (min_memberships + min_non_memberships) / 
                    (2.0 * (max_memberships + max_non_memberships)))
            )
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_1_3:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        max_memberships = np.maximum(A.membership_values, B.membership_values)
        max_non_memberships = np.maximum(A.non_membership_values, B.non_membership_values)
        return 1 - np.power(
            np.sum(
                np.power(
                    (delta_memberships + delta_non_memberships) /
                    (max_memberships + max_non_memberships), p
                )
            ) / n, 1.0 / p
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_1_4:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        max_memberships = np.maximum(A.membership_values, B.membership_values)
        max_non_memberships = np.maximum(A.non_membership_values, B.non_membership_values)
        return (
            (n - np.sum(
                (delta_memberships + delta_non_memberships) /
                (max_memberships + max_non_memberships)
            )) / 
            (n + np.sum(
                (delta_memberships + delta_non_memberships) /
                (max_memberships + max_non_memberships)
            ))
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_2_1:
        n = len(A)
        return (1.0 / n) * np.sum(
            (1.0 - np.abs(A.non_membership_values - B.non_membership_values)) /
            (1.0 + np.abs(A.membership_values - B.membership_values))
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_2_2:
        return (
            np.sum(1.0 - np.abs(A.non_membership_values - B.non_membership_values)) /
            np.sum(1.0 + np.abs(A.membership_values - B.membership_values))
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_2_3:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        return 1.0 - np.power(
            np.sum(np.power(
                (delta_memberships + delta_non_memberships) / 2.0, p
            )) / n, 1.0 / p
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_2_4:
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        return (
            np.sum(2.0 - delta_memberships - delta_non_memberships) / 
            np.sum(2.0 + delta_memberships + delta_non_memberships)
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_3_1:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        return (
            (
                np.e ** 2.0 - 
                np.power(
                    np.sum(
                        np.power(delta_memberships, p)
                    ) / n, 1.0 / p
                ) -
                np.power(
                    np.sum(
                        np.power(delta_non_memberships, p)
                    ) / n, 1.0 / p
                ) -
                1.0
            ) /
            (np.e ** 2.0 - 1.0)
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_3_2:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        return 1.0 - (
            (
                np.power(
                    np.power(
                        np.sum(np.power(delta_memberships, p)) / n, p
                    ), u
                ) + 
                np.power(
                    np.power(
                        np.sum(np.power(delta_non_memberships, p)) / n, p
                    ), v
                )
            ) / 2.0
        )
    elif similarity_type == DENG_JIANG_FU_MONOTONIC_TYPE_3_3:
        n = len(A)
        delta_memberships = np.abs(A.membership_values - B.membership_values)
        delta_non_memberships = np.abs(A.non_membership_values - B.non_membership_values)
        return 1.0 - (
            np.sin(np.pi / 2.0 * np.power(
                np.sum(np.power(delta_memberships, p)) / n, 1.0 / p
            )) + 
            np.sin(np.pi / 2.0 * np.power(
                np.sum(np.power(delta_non_memberships, p)) / n, 1.0 / p
            )) 
        ) / 2.0
    else:
        raise ValueError("similarity_type parameter must be DENG_JIANG_FU_MONOTONIC_TYPE_1_1...4, DENG_JIANG_FU_MONOTONIC_TYPE_2_1...4 or DENG_JIANG_FU_MONOTONIC_TYPE_3_1...3.")


def nguyen(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet):
    """ Similarity proposed by H. Nguyen, from the related article: 
    "A novel similarity/dissimilarity measure for intuitionistic fuzzy sets and its application in pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)

    def k_single(m, v, p):
        return np.sqrt(
            np.power(m, 2.0) +
            np.power(v, 2.0) +
            np.power(1.0 - p, 2.0)
        )
    def k(set: IntuitionisticFuzzySet):
        return 1.0 / (len(set) * np.sqrt(2)) * np.sum(np.sqrt(
            np.power(set.membership_values, 2.0) +
            np.power(set.non_membership_values, 2.0) +
            np.power(1.0 - set.hesitation_degrees, 2.0)
        ))
    def membership_knowledge(set: IntuitionisticFuzzySet):
        memberships, non_memberships, hestitations = set.membership_values, set.non_membership_values, set.hesitation_degrees
        K = 1.0 / (len(set) * np.sqrt(2)) * np.array([
            k_single(m, v, p) if m >= v else -k_single(m, v, p)
            for m, v, p in zip(memberships, non_memberships, hestitations)
        ])
        return 1.0 / len(set) * np.sum(K)
        
    _Ka = membership_knowledge(A)
    _Kb = membership_knowledge(B)
    Ka = k(A)
    Kb = k(B)
    return 1.0 - np.abs(Ka - Kb) if _Ka * _Kb >= 0 else np.abs(Ka - Kb) - 1.0


def chen_cheng_lan(A: IntuitionisticFuzzySet, B: IntuitionisticFuzzySet, weights=None):
    """ Similarity proposed by S.M. Chen, S.H. Cheng, T.-C. Lan, from the related article: 
    "A novel similarity measure between intuitionistic fuzzy sets based on the centroid points of transformed fuzzy numbers with applications to pattern recognition"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)
    check_weights(weights, len(A))

    if weights is None:
        weights = np.full(len(A), 1.0 / len(A))
    
    delta_memberships = A.membership_values - B.membership_values
    delta_non_memberships = A.non_membership_values - B.non_membership_values
    return np.sum(
        weights * (
            1.0 - 
            np.abs(2.0 * delta_memberships - delta_non_memberships) / 3.0 * (1.0 - (A.hesitation_degrees + B.hesitation_degrees) / 2.0) -
            np.abs(2.0 * delta_non_memberships - delta_memberships) / 3.0 * ((A.hesitation_degrees + B.hesitation_degrees) / 2.0)
        )
    )


def muthukumar_krishnanb(A: FuzzySet, B: FuzzySet, weights=None):
    """ Similarity proposed by P. Muthukumar, G. S. S. Krishnan, from the related article: 
    "A similarity measure of intuitionistic fuzzy soft sets and itsapplication in medical diagnosis"

    Parameters
    ----------
    A : FuzzySet
        A fuzzy set.
    B : FuzzySet
        A fuzzy set.
    weights : list of floats
        List of weights for each membership/non-membership value.
        
    Returns
    -------
    numpy.float64
        The similarity between the two sets provided.
    """
    check_sets_cardinality(A, B)

    if weights is None:
        return (
            np.sum(
                A.membership_values * B.membership_values + 
                A.non_membership_values * B.non_membership_values
            ) / 
            np.sum(
                np.maximum(np.power(A.membership_values, 2.0), np.power(B.membership_values, 2.0)) + 
                np.maximum(np.power(A.non_membership_values, 2.0), np.power(B.non_membership_values, 2.0))
            )
        )
    else:
        check_weights(weights, len(A))
        return (
            np.sum(
                weights * 
                (A.membership_values * B.membership_values + 
                A.non_membership_values * B.non_membership_values)
            ) / 
            np.sum(
                np.maximum(np.power(A.membership_values, 2.0), np.power(B.membership_values, 2.0)) + 
                np.maximum(np.power(A.non_membership_values, 2.0), np.power(B.non_membership_values, 2.0))
            )
        ) / np.sum(weights)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

try:
    import builtins
except ImportError:
    # Python 2 compat: just to be able to declare that Python >=3.5 is needed.
    import __builtin__ as builtins

import fsmpy

VERSION = fsmpy.__version__

CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development",
    "Topic :: Utilities",
    "Operating System :: Microsoft",
    "Operating System :: MacOS",
    "Operating System :: POSIX :: Linux",
]

# This call to setup() does all the work
setup(
    name="fuzzy-set-measures",
    version=VERSION,
    description="Generalized measures for application in fuzzy set theory.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="",
    author="Machine Learning & Vision Research Group",
    license="BSD 3-Clause License",
    classifiers=CLASSIFIERS,
    packages=[
        "fsmpy", "fsmpy.utils"
    ],
    include_package_data=True,
    install_requires=[
        "numpy >= 1.14.6",
        "scikit-learn >=0.24.2",
        "tqdm >= 4.62.1",
    ],
)


from .sets import FuzzySet


def load_patients_diagnoses():
    """ Returns diagnoses and patients data for the medical diagnosis application.

    The data returned are consisted of FuzzySets representing the symptoms each
    patient has and the symptoms of each diagnosis.
    The symptoms include the following:
    **Temperature**, **Headache**, **Stomach paint**, **Cough** and **Chest paint**.

        
    Returns
    -------
    diagnoses : list[FuzzySet]
        List of FuzzySets representing each giadnosis: 
        **Viral fever**, **Malaria**, **Typhoid**, **Stomach problem** and **Chest problem**. 
    patients : list[FuzzySet]
        List of FuzzySets representing each patient's symptoms:
        Al, Bob, Joe and Ted.
    
    Examples
    --------
    Diagnoses:

    >>> viral_fever = FuzzySet([0.4, 0.3, 0.1, 0.4, 0.1], [0.0, 0.5, 0.7, 0.3, 0.7])
    >>> malaria = FuzzySet([0.7, 0.2, 0.0, 0.7, 0.1], [0.0, 0.6, 0.9, 0.0, 0.8])
    >>> typhoid = FuzzySet([0.3, 0.6, 0.2, 0.2, 0.1], [0.3, 0.1, 0.7, 0.6, 0.9])
    >>> stomach_problem = FuzzySet([0.1, 0.2, 0.8, 0.2, 0.2], [0.7, 0.4, 0.0, 0.7, 0.7])
    >>> chest_problem = FuzzySet([0.1, 0.0, 0.2, 0.2, 0.8], [0.8, 0.8, 0.8, 0.8, 0.1])

    Patients:
    
    >>> al = FuzzySet([0.8, 0.6, 0.2, 0.6, 0.1], [0.1, 0.1, 0.8, 0.1, 0.6])
    >>> bob = FuzzySet([0.0, 0.4, 0.6, 0.1, 0.1], [0.8, 0.4, 0.1, 0.7, 0.8])
    >>> joe = FuzzySet([0.8, 0.8, 0.0, 0.2, 0.0], [0.1, 0.1, 0.6, 0.7, 0.5])
    >>> ted = FuzzySet([0.6, 0.5, 0.3, 0.7, 0.3], [0.1, 0.4, 0.4, 0.2, 0.4])
    """
    # diagnosis = [temperature, headache, stomach pain, cough, chest paint]
    viral_fever = FuzzySet([0.4, 0.3, 0.1, 0.4, 0.1],
                           [0.0, 0.5, 0.7, 0.3, 0.7])
    malaria = FuzzySet([0.7, 0.2, 0.0, 0.7, 0.1], [0.0, 0.6, 0.9, 0.0, 0.8])
    typhoid = FuzzySet([0.3, 0.6, 0.2, 0.2, 0.1], [0.3, 0.1, 0.7, 0.6, 0.9])
    stomach_problem = FuzzySet([0.1, 0.2, 0.8, 0.2, 0.2], [
                               0.7, 0.4, 0.0, 0.7, 0.7])
    chest_problem = FuzzySet([0.1, 0.0, 0.2, 0.2, 0.8], [
                             0.8, 0.8, 0.8, 0.8, 0.1])
    diagnoses = [viral_fever, malaria, typhoid, stomach_problem, chest_problem]

    # patient = membership values for each characteristic, non-membership values for each characteristic
    al = FuzzySet([0.8, 0.6, 0.2, 0.6, 0.1], [0.1, 0.1, 0.8, 0.1, 0.6])
    bob = FuzzySet([0.0, 0.4, 0.6, 0.1, 0.1], [0.8, 0.4, 0.1, 0.7, 0.8])
    joe = FuzzySet([0.8, 0.8, 0.0, 0.2, 0.0], [0.1, 0.1, 0.6, 0.7, 0.5])
    ted = FuzzySet([0.6, 0.5, 0.3, 0.7, 0.3], [0.1, 0.4, 0.4, 0.2, 0.4])
    patients = [al, bob, joe, ted]

    return diagnoses, patients
    